<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

namespace core_backup\hook;

use stdClass;
use restore_plan;

/**
 * Hook used to allow interaction with the copy task, before the actual task execution takes place.
 *
 * @package core_backup
 * @copyright 2024 Monash University (https://www.monash.edu)
 * @license http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
#[\core\attribute\label('Use to interact with the copy course task before the actual execution takes place.')]
#[\core\attribute\tags('backup')]
final class before_copy_course_execute {

    /** @var restore_plan */
    public readonly restore_plan $plan;

    /** @var stdClass */
    public readonly stdClass $copyinfo;

    /**
     * Hook constructor.
     *
     * @param restore_plan $plan
     * @param stdClass $copyinfo
     */
    public function __construct(restore_plan $plan, stdClass $copyinfo) {
        $this->plan = $plan;
        $this->copyinfo = $copyinfo;
    }
}
